# prast.py (encoding: utf-8) ver 0.9.1 | 2018/11/05
  # Copyright (C) T. Yoshiizumi, 2018 All rights reserved.
  # 定番の単純な集計を行うためのPrastクラスを定義
import os, sys, platform
import pandas as pd
import scipy.stats as stats
import numpy as np
import re
from platform import python_version
from collections import OrderedDict

# general variables
__version__ = "0.9.1"
mrange = None

# general methods
def print2(s):
    if int(python_version()[0]) < 3:
        cmd = "print >> sys.stderr, s"
        exec(cmd)
    else:
        cmd = "print(s, end = \"\\n\", file = sys.stderr)"
        dct = {'s': s}
        exec(cmd, globals(), dct)


def chisq_test(dtf):
    if not type(dtf) == pd.DataFrame:
        return None
    clist = list(dtf.columns)
    ilist = list(dtf.index)
    if len(clist) < 2 or len(ilist) < 2:
        print2(u"Error: in chisq_test(), DataFrameの列または行の数が２未満!")
        print2(dtf)
        print2('')
        return None
    dtf.columns = clist
    dtf.index = ilist
    tbl = np.array(dtf)
    chi2, p, df, expected = stats.chi2_contingency(tbl)
    expected = np.array(expected)
    res = tbl - expected
    residuals = res / np.sqrt(expected)
    tsum = tbl.sum()*1.0
    celm = 1.0 - np.sum(tbl, axis=0) / tsum
    relm = 1.0 - np.sum(tbl, axis=1) / tsum
    mouter = np.outer(relm, celm)
    stdres = residuals / np.sqrt(mouter)
    dod = OrderedDict()
    dod['chi2'] = chi2
    dod['df'] = df
    dod['p_value'] = p
    dod['observed'] = dtf
    dod['expected'] = pd.DataFrame(expected, index=ilist, columns=clist)
    dod['residuals'] = pd.DataFrame(residuals, index=ilist, columns=clist)
    dod['stdres'] = pd.DataFrame(stdres, index=ilist, columns=clist)
    ttl = expected.size
    less5 = len(np.where(expected < 5)[0])
    lpct = less5*1.0 / ttl*1.0 * 100.0
    if lpct >= 20.0:
        ws = u"警告: '期待度数が5未満のセルの割合' >= 20% | " + \
            u"%.1f%%(%d/%d)" % (lpct, less5, ttl)
        print2(ws)
        dod['warning'] = ws
    dod['cramer'] = np.sqrt(chi2 / (tsum*(min(tbl.shape)-1)))
    return dod

def var_test(a, b, alternative="two-sided"):
    alt_arg = alternative
    aa = np.array(a)
    bb = np.array(b)
    aa = aa[~np.isnan(aa)]
    bb = bb[~np.isnan(bb)]
    aa_var = np.var(aa, ddof=1)  # aの不偏分散
    bb_var = np.var(bb, ddof=1)  # bの不偏分散
    aa_df = len(aa) - 1  # aの自由度
    bb_df = len(bb) - 1  # bの自由度
    f = aa_var / bb_var  # F比の値
    pval1 = stats.f.cdf(f, aa_df, bb_df)  # 片側検定のp値・その１
    pval2 = stats.f.sf(f, aa_df, bb_df)  # 片側検定のp値・その２
    pval3 = min(pval1, pval2) * 2  # 両側検定のp値
    if alt_arg == 'less':
        alt = u"対立仮説: F比が 1.0 より小さい."
        pval = pval1
    elif alt_arg == 'greater':
        alt = u"対立仮説: F比が 1.0 より大きい."
        pval = pval2
    elif alt_arg == 'two-sided':
        alt = u"対立仮説: F比が 1.0 ではない."
        pval = pval3
    else:
        alt = "3 p_values(less, greater, two-sided)"
        pval = [pval1, pval2, pval3]
    dod = OrderedDict()
    dod['f'] = f
    dod['p_value'] = pval
    dod['alt'] = alt + " (%s)" % alternative
    dod['var'] = [aa_var, bb_var]
    dod['df'] = [aa_df, bb_df]
    return dod

def t_test(a, b, alternative="two-sided", paired=False, equal_var=False):
    eqv = equal_var
    alt_arg = alternative
    aa = np.array(a)
    bb = np.array(b)
    aa = aa[~np.isnan(aa)]
    bb = bb[~np.isnan(bb)]
    aa_mean = np.mean(aa)  # aの平均値
    bb_mean = np.mean(bb)  # bの平均値
    if not paired:
        t, p = stats.ttest_ind(aa, bb, equal_var=eqv)
    else:
        if not len(aa) == len(bb):
            print2(u"Error: ２つの配列の要素個数が不一致!")
            return None
        t, p = stats.ttest_rel(aa, bb)
    pval3 = p
    pval2 = p / 2.0
    pval1 = 1.0 - pval2
    if t < 0.0:
        w = pval2
        pval2 = pval1
        pval1 = w
    if alt_arg == 'less':
        alt = u"対立仮説: 平均値の差が 0 より小さい."
        pval = pval1
    elif alt_arg == 'greater':
        alt = u"対立仮説: 平均値の差が 0 より大きい."
        pval = pval2
    elif alt_arg == 'two-sided':
        alt = u"対立仮説: 平均値に差がある."
        pval = pval3
    else:
        alt = "3 p_values(less, greater, two-sided)"
        pval = [pval1, pval2, pval3]
    dod = OrderedDict()
    dod['t']= t
    dod['p_value'] = pval
    dod['alt'] = alt + " (%s)" % alternative
    dod['mean'] = [aa_mean, bb_mean]
    dod['difference'] = aa_mean -bb_mean
    dod['count'] = [len(aa), len(bb)]
    return dod

def join_table(dtf1, dtf2, fmt="%s(%s)"):
    ilist = list(dtf1.index)
    dtfx = pd.DataFrame()
    for cn in dtf1.columns:
        s1 = dtf1[cn]
        s2 = dtf2[cn]
        l = list()
        for i in range(len(s1)):
            v = fmt % (s1[i], s2[i])
            l.append(v)
        dtfx[cn] = pd.Series(l, index=ilist)
    return dtfx

def set_encoding(enc):
    if int(python_version()[0]) < 3:
        reload(sys)
        sys.setdefaultencoding(enc)

def single2multi(dtf, mcolumn, sep=",", value=1, na=np.NaN, clist=None):
    ilist = list(dtf.index)
    mlist = list(dtf[mcolumn])
    key_dct = {}
    ptn1 = re.compile(r"^\s+")
    ptn2 = re.compile(r"\s+$")
    for i in range(len(mlist)):
        elm = mlist[i]
        dct = {}
        if not elm == elm:
            mlist[i] = dct
            continue
        elm = elm.split(sep)
        for k in range(len(elm)):
            elm[k] = re.sub(ptn1, "", elm[k])
            elm[k] = re.sub(ptn2, "", elm[k])
        for key in elm:
            key_dct[key] = 1
            dct[key] = 1
        mlist[i] = dct
    keys1 = list(key_dct.keys())
    if type(clist) == list:
        if sorted(clist) == sorted(keys1):
            keys1 = clist
        else:
            print2(u"Warning: clist %s が不適合!" % clist)
    dtf2 = pd.DataFrame()
    for i in range(len(mlist)):
        dct = mlist[i]
        keys2 = dct.keys()
        w = []
        for key in keys1:
            if key in keys2:
                w.append(value)
            else:
                w.append(na)
        dtf2[ilist[i]] = pd.Series(w)
    dtf2.index = keys1
    dtf2 = dtf2.T
    dtf = pd.concat([dtf, dtf2], axis=1)
    global mrange
    mrange = keys1
    return dtf

def multi2single(dtf, clist, mcolumn, sep=",", value=1):
    dtfr = dtf.loc[:, clist].T
    mlist = list()
    for col in dtfr.columns:
        ser = dtfr[col]
        ser = ser.apply(lambda x: True if x == value else None).dropna()
        ilist = list(ser.index)
        istr = sep.join(map(str, ilist))
        mlist.append(istr)
    ser = pd.Series(mlist)
    ser.index = list(dtf.index)
    dtf[mcolumn] = ser
    return dtf


# class definition
class Prast:
    etoj1 = {'count': u'度数', 'mean': u'平均', 'std': 'std',
        'min': u'最小値', 'max': u'最大値'}
    etoj2 = {'count': u'度数', 'unique': u'種類数',
        'top': u'最頻値', 'freq': u'最頻度数'}
    etoj3 = {'count': u'度数', 'ratio': u'割合'}
    dround = {'count': 0, 'mean': 1, 'std': 2, 'min': 1, '%': 1, 'max': 1}
    cround = 1
    if int(python_version()[0]) < 3:
        num_types = [int, long, float]
    else:
        num_types = [int, float]

    def __init__(self, dtf, columns=None, indexes=None, encoding=None):
        self.__encoding = 'utf-8'
        self.__dtf = None
        self.__columns = None
        self.__indexes = None
        self.put_encoding(encoding)
        self.put_dtf(dtf)  ## set DataFrame
        self.put_columns(columns)  ## set columns' correspondent table
        self.put_indexes(indexes)  ## set indexes' correspondent table

    def get_encoding(self):
        return self.__encoding

    def put_encoding(self, encoding=None):
        if not encoding is None:
            self.__encoding = encoding
        set_encoding(self.__encoding)
        return self.__encoding

    ## private methods
    def __csv_reader(self, s):
        delimiter = ","
        if "\t" in s:
            delimiter = "\t"
        a = s.split("\n")
        while "" in a:
            a.remove("")
        aa = list()
        for s2 in a:
            a2 = s2.split(delimiter)
            aa.append(a2)
        return aa

    def __str2ld(self, s):
        try:
            s = re.sub(r"\A\n+", "", s)
        except:
            return s
        s = re.sub(r"\n+\Z", "", s)
        x = self.__csv_reader(s)
        if not type(x) == list:
            return s
        if type(x[0]) == list:
            if len(x[0]) == 2:
                x = OrderedDict(x)
            elif len(x) == 1:
                x = x[0]
            else:
                x2 = list()
                for i in range(len(x)):
                    x2.append(x[i][0])
                x = x2
        return x

    def __str2od(self, str):
        try:
            str = re.sub(r"\A\n+", "", str)
        except:
            return str
        str = re.sub(r"\n+\Z", "", str)
        str = re.sub(r"\n\n\n+", "\n\n", str)
        a1 = str.split("\n\n")
        dod = OrderedDict()
        count = 0
        for s in a1:
            a2 = self.__csv_reader(s)
            if len(a2) == 0:
                continue
            if len(a2[0]) == 1:
                k = a2.pop(0)[0]
                dod[k] = OrderedDict(a2)
            else:
                count = count + 1
                k = "_key%d" % count
                dod[k] = OrderedDict(a2)
        return dod

    def __get_inames(self, key, all=False):
        if not type(self.__indexes) == OrderedDict:
            return None
        if not key in self.__indexes.keys():
            return None
        inames = self.__indexes[key].copy()
        if all and (not 'all' in inames):
            inames['_all'] = u'全体'
        return inames

    def __get_cnames(self):
        if not type(self.__columns) == OrderedDict:
            return None
        cnames = self.__columns.copy()
        return cnames

    def __prod(self, blist, slist):  # product of two lists
        blist = list(blist)
        slist = list(slist)
        if type(slist[0]) in self.num_types:
            slist = self.__to_str(slist)
        ilist = []
        for elm in blist:
            if elm in slist:
                ilist.append(elm)
        return ilist

    def __cross_base(self, gkey1, gkey2):
        dtf = pd.DataFrame()
        dtf[gkey1] = self.__dtf[gkey1]
        dtf[gkey2] = self.__dtf[gkey2]
        ser1 = dtf[gkey1] == dtf[gkey1]
        ser2 = dtf[gkey2] == dtf[gkey2]
        dtf = dtf[ser1 & ser2]
        dtf['_tmp'] = dtf[gkey1]
        dtfx = pd.pivot_table(dtf, values=['_tmp'], index=gkey1,
            columns=gkey2, aggfunc='count')
        dtfx = dtfx['_tmp']
        dtfx.index = list(dtfx.index)
        cl = list(dtfx.columns)
        if type(cl[0]) in self.num_types:
            dtfx.columns = self.__to_str(cl)
        il = list(dtfx.index)
        if type(il[0]) in self.num_types:
            dtfx.index = self.__to_str(il)
        for col in dtfx.columns:
            dtfx[col] = dtfx[col].fillna(0)
            dtfx[col] = dtfx[col].astype(int)
        inames = self.__get_inames(gkey1)
        if type(inames) == OrderedDict:
            ilist = self.__prod(inames.keys(), dtfx.index)
            dtfx = dtfx.loc[ilist]
            dtfx.rename(index=inames, inplace=True)
        cnames = self.__get_inames(gkey2)
        if type(cnames) == OrderedDict:
            clist = self.__prod(cnames.keys(), dtfx.columns)
            dtfx = dtfx.loc[:, clist]
            dtfx.rename(columns=cnames, inplace=True)
        return dtfx

    def __str_type(self, x):
        return str(x) in ['object', 'category']

    def __to_str(self, wlist):  # each element of a list to string
        wlist2 = list()
        ptn = re.compile(r'\.0+$')
        for i in range(len(wlist)):
            w = str(wlist[i])
            w = re.sub(ptn, '', w)  # to integer string
            wlist2.append(w)
        return wlist2

    def __key_str(self, xd):
        if not type(xd) in [dict, OrderedDict]:
            return xd
        new_od = OrderedDict()
        ptn = re.compile(r'\.0+$')
        for key in xd.keys():
            key2 = str(key)
            key2 = re.sub(ptn, '', key2)
            new_od[key2] = self.__key_str(xd[key])
        return new_od

    def __unique(self, gkey):
        dser = self.__dtf.dtypes
        ser = self.__dtf[gkey].dropna()
        if not str(dser[gkey]) == 'object':
            ser = ser.astype(str)
            ptn = re.compile(r'\.0+$')
            ser = ser.apply(lambda x: re.sub(ptn, '', x))
        return list(ser.unique())

    ## public methods
    def get_dtf(self):
        return self.__dtf
    gd = get_dtf

    def get_columns(self):
        return self.__columns

    def get_indexes(self):
        return self.__indexes

    def put_dtf(self, dtf):
        if dtf is None:
            self.__dtf = None
            return None
        if not type(dtf) == pd.DataFrame:
            print2("Error: 'dtf' should be pandas.DataFrame!")
            return self.__dtf
        dser = dtf.dtypes
        cl = list()
        for cname in list(dser.index):
            t = str(dser[cname])
            if t == 'category':
                cl.append(cname)
        if len(cl) == 0:
            self.__dtf = dtf
        else:
            self.__dtf = dtf.copy()
            for cname in cl:
                self.__dtf[cname] = list(self.__dtf[cname])
        return self.__dtf
    pd = put_dtf

    def put_columns(self, columns):
        if columns is None:
            self.__columns = None
            return None
        w = self.read_text(columns)
        if w is None:
            w = self.__key_str(columns)
        if not w is None:
            w = self.__str2ld(w)
            if type(w) == OrderedDict:
                if not type(self.__columns) == OrderedDict:
                    self.__columns = OrderedDict()
                self.__columns.update(w)
        if type(self.__columns) == OrderedDict and len(self.__columns) == 0:
            self.__columns = None
        return self.__columns

    def put_indexes(self, indexes):
        if indexes is None:
            self.__indexes = None
            return None
        w = self.read_text(indexes)
        if w is None:
            w = self.__key_str(indexes)
        if not w is None:
            wod = self.__str2od(w)
            if type(wod) == OrderedDict:
                if not type(self.__indexes) == OrderedDict:
                    self.__indexes = OrderedDict()
                self.__indexes.update(wod)
        if type(self.__indexes) == OrderedDict and len(self.__indexes) == 0:
            self.__indexes = None
        return self.__indexes

    def copy(self):
        new_dtf = None
        if not self.__dtf is None:
            new_dtf = self.__dtf.copy()
        new_columns = None
        if not self.__columns is None:
            new_columns = self.__columns.copy()
        new_indexes = None
        if not self.__indexes is None:
            new_indexes = self.__indexes.copy()
        new_encoding = self.__encoding
        new_obj = Prast(new_dtf, new_columns, new_indexes, new_encoding)
        return new_obj

    def fillna(self, clist=None, naval=None, exclude=None):
        dser = self.__dtf.dtypes
        if clist is None:
            clist = list()
            for cname in dser.index:
                if self.__str_type(dser[cname]):
                    clist.append(cname)
        elif not type(clist) == list:
            clist = [clist]
        if exclude is None:
            exclude = list()
        elif not type(exclude) == list:
            exclude = [exclude]
        for cname in clist:
            if cname in exclude:
                continue
            nv = naval
            if nv is None:
                nv = '_na'
            ser = self.__dtf[cname].fillna(nv)
            self.__dtf[cname] = ser
        return self.__dtf

    def recona(self, clist=None, naval=None, exclude=None):
        dser = self.__dtf.dtypes
        if clist is None:
            clist = list()
            for cname in dser.index:
                if self.__str_type(dser[cname]):
                    clist.append(cname)
        elif not type(clist) == list:
            clist = [clist]
        if exclude is None:
            exclude = list()
        elif not type(exclude) == list:
            exclude = [exclude]
        for cname in clist:
            if cname in exclude:
                continue
            nv = naval
            if nv is None:
                nv = '_na'
            ser = self.__dtf[cname]
            self.__dtf[cname] = ser.replace(nv, np.NaN)
        return self.__dtf

    def read_text(self, filename, charset=''):
        try:
            flag = os.path.exists(filename)
        except:
            flag = False
        if not flag:
            return None
        if charset == '':
            charset = self.__encoding
        if int(python_version()[0]) < 3:
            f = open(filename, mode='r')
            data = f.read()
            f.close()
            data = unicode(data, charset)
        else:
            f = open(filename, mode='rt', encoding='cp932')
            data = f.read()
            f.close()
        return data

    def describe_arrange(self, dtfz):
        if dtfz.empty:
            return dtfz
        dtfz = dtfz.T
        if 'mean' in dtfz.columns and \
            type(self.dround) in [dict, OrderedDict]:
            for cname in dtfz.columns:
                if cname in self.dround.keys():
                    fsize = self.dround[cname]
                elif '%' in cname:
                    fsize = self.dround['%']
                else:
                    fsize = 1
                if fsize == 0:
                    dtfz[cname] = dtfz[cname].astype(int)
                else:
                    dtfz[cname] = dtfz[cname].round(fsize)
            if type(self.etoj1) in [dict, OrderedDict]:
                dtfz.rename(columns=self.etoj1, inplace=True)
        else:
            if type(self.etoj2) in [dict, OrderedDict]:
                dtfz.rename(columns=self.etoj2, inplace=True)
        return dtfz

    def describe(self, clist, gkey=None, ddof=True, percentiles=[.25,.5,.75]):
        if not type(clist) == list:
            clist = [clist]
        cnames = self.__get_cnames()
        if not gkey is None:
            inames = self.__get_inames(gkey, True)
            if not inames is None:
                ilist = list(inames.keys())
            else:
                ilist = self.__unique(gkey)
                ilist.append('_all')
            ser = self.__dtf[gkey] == self.__dtf[gkey]
            dtf = self.__dtf[ser]
            dgb = dtf.groupby(gkey)  # groupping by gkey
            dct = dict()
            for g_name, g_dtf in dgb:  # dgb to dict
                dct[g_name] = g_dtf
            dct['_all'] = dtf  # append '_all'
            dct = self.__key_str(dct)
            dod = OrderedDict()
            for cname in clist:
                dtfx = pd.DataFrame()
                for idx in ilist:
                    if not idx in dct.keys():
                        continue
                    if not cname in dct[idx].keys():
                        continue
                    ser = dct[idx][cname].describe(percentiles=percentiles)
                    if not 'mean' in ser:
                        continue
                    if ddof == False:
                        ser['std'] = dct[idx][cname].std(ddof=False)
                    dtfx[idx] = ser
                dtfx = self.describe_arrange(dtfx)
                if type(inames) == OrderedDict:
                    dtfx.rename(index = inames, inplace = True)
                else:
                    dtfx.rename(index = {'_all': u'全体'}, inplace = True)
                ckey = cname
                if type(cnames) == OrderedDict:
                    ckey = cnames[cname]
                if dtfx.empty:
                    print2(u"'%s': 数値データでないため処理対象外." % cname)
                else:
                    dod[ckey] = dtfx
            return dod
        else:  # if gkey is None
            dod = OrderedDict()
            dtfx = pd.DataFrame()
            dtfy = pd.DataFrame()
            for cname in clist:
                if not cname in self.__dtf.columns:
                    print2(u"'%s': 該当の列名なし." % cname)
                    continue
                ckey = cname
                if type(cnames) == OrderedDict:
                    ckey = cnames[cname]
                ser = self.__dtf[cname].describe(percentiles=percentiles)
                if 'mean' in ser:
                    if ddof == False:
                        ser['std'] = self.__dtf[cname].std(ddof=False)
                    dtfx[ckey] = ser
                else:  # if not 'mean' in ser
                    inames = self.__get_inames(cname)
                    if type(inames) == OrderedDict:
                        ser['top'] = inames[ser['top']]
                    dtfy[ckey] = ser
            if not dtfx.empty:
                dtfx = self.describe_arrange(dtfx)
                if type(cnames) == OrderedDict:
                    dtfx.rename(index=cnames, inplace=True)
                dod[u'数値型'] = dtfx
            if not dtfy.empty:
                dtfy = self.describe_arrange(dtfy)
                if type(cnames) == OrderedDict:
                    dtfy.rename(index=cnames, inplace=True)
                dod[u'頻度型'] = dtfy
            return dod

    def shapiro(self, clist, gkey=None):
        if not type(clist) == list:
            clist = [clist]
        cnames = self.__get_cnames()
        if not gkey is None:
            inames = self.__get_inames(gkey, True)
            if not inames is None:
                ilist = list(inames.keys())
            else:
                ilist = self.__unique(gkey)
                ilist.append('_all')
            ser = self.__dtf[gkey] == self.__dtf[gkey]
            dtf = self.__dtf[ser]
            dgb = dtf.groupby(gkey)  # groupping by gkey
            dct = dict()
            for g_name, g_dtf in dgb:  # dgb to dict
                dct[g_name] = g_dtf
            dct['_all'] = dtf  # append '_all'
            dct = self.__key_str(dct)
            dod = OrderedDict()
            for cname in clist:
                dtfx = pd.DataFrame()
                for idx in ilist:
                    if not idx in dct.keys():
                        continue
                    if not cname in dct[idx].keys():
                        continue
                    ser = dct[idx][cname].dropna().reset_index(drop=True)
                    try:
                        w, p = stats.shapiro(ser)
                        n = 'yes'
                        if p < 0.05:
                            n = 'no'
                    except:
                        w, p, n = [np.NaN, np.NaN, np.NaN]
                    ser = pd.Series([w, p, n],
                        index=['w', 'p_value', u'正規分布'])
                    dtfx[idx] = ser
                if type(inames) == OrderedDict:
                    dtfx.rename(columns = inames, inplace = True)
                else:
                    dtfx.rename(columns = {'_all': u'全体'}, inplace = True)
                ckey = cname
                if type(cnames) == OrderedDict:
                    ckey = cnames[cname]
                if dtfx.empty:
                    print2(u"'%s': 数値データでないため処理対象外." % cname)
                else:
                    dod[ckey] = dtfx.T
            return dod
        else:  # if gkey is None
            dod = OrderedDict()
            dtfx = pd.DataFrame()
            for cname in clist:
                if not cname in self.__dtf.columns:
                    print2(u"'%s': 該当の列名なし." % cname)
                    continue
                ckey = cname
                if type(cnames) == OrderedDict:
                    ckey = cnames[cname]
                ser = self.__dtf[cname].dropna().reset_index(drop=True)
                try:
                    w, p = stats.shapiro(ser)
                    n = 'yes'
                    if p < 0.05:
                        n = 'no'
                except:
                    w, p, n = [np.NaN, np.NaN, np.NaN]
                ser = pd.Series([w, p, n],
                    index=['w', 'p_value', u'正規分布'])
                dtfx[ckey] = ser
            if not dtfx.empty:
                dod['shapiro'] = dtfx.T
            return dod

    def count(self, clist=None, asc=None, exclude=None):
        cnames = self.__get_cnames()
        dod = OrderedDict()
        dser = self.__dtf.dtypes
        if clist is None:
            clist = list()
            for cname in dser.index:
                if self.__str_type(dser[cname]):
                    clist.append(cname)
        elif not type(clist) == list:
            clist = [clist]
        if exclude is None:
            exclude = list()
        elif not type(exclude) == list:
            exclude = [exclude]
        for cname in clist:
            if cname in exclude:
                continue
            inames = self.__get_inames(cname, True)
            ser = self.__dtf[cname] == self.__dtf[cname]
            dtf = self.__dtf[ser]
            ser = dtf[cname]
            dtf = dtf.assign(count = ser)
            dtf2 = pd.pivot_table(dtf, values=['count'],
                index=cname, aggfunc='count')
            il = list(dtf2.index)
            if type(il[0]) in self.num_types:
                dtf2.index = self.__to_str(il)
            if asc in [True, False]:
                dtf2 = dtf2.sort_values('count', ascending=asc)
            elif type(inames) == OrderedDict:
                ilist = self.__prod(inames.keys(), dtf2.index)
                dtf2 = dtf2.loc[ilist]
            dtf2 = pd.concat([dtf2, pd.DataFrame(dtf2.sum(axis=0),
                columns=['_all']).T])
            ser = dtf2['count']
            ser = ser / ser[-1] * 100.0
            if type(self.cround) == int:
                ser = ser.round(self.cround)
            dtf2['ratio'] = ser
            if type(self.etoj3) in [dict, OrderedDict]:
                dtf2.rename(columns=self.etoj3, inplace=True)
            if type(inames) == OrderedDict:
                dtf2.rename(index=inames, inplace=True)
            else:
                dtf2.rename(index={'_all': u'全体'}, inplace=True)
            key = cname
            if type(cnames) == OrderedDict and cname in cnames:
                key = cnames[cname]
            dtf2.columns.name = key
            dod[key] = dtf2
        return dod

    def cut(self, col, new_col, bins, labels=None):
        if type(new_col) in [dict, OrderedDict]:
            new_col1 = list(new_col.keys())[0]
            new_col2 = list(new_col.values())[0]
        elif type(new_col) == list and len(new_col) == 2:
            new_col1 = new_col[0]
            new_col2 = new_col[1]
        elif type(new_col) == list:
            new_col1 = new_col[0]
            new_col2 = new_col[0]
        else:
            new_col1 = new_col
            new_col2 = new_col
        labels1 = ['1']
        labels2 = [u"～%s" % bins[1]]
        for i in range(1, len(bins)-1):
            labels1.append(str(i+1))
            labels2.append(u"%s～%s" % (bins[i], bins[i+1]))
        labels2[-1] = u"%s～" % bins[-2]
        if not labels is None:
            labels2 = labels
        cdata = pd.cut(self.__dtf[col], bins, right=False, labels=labels1)
        self.__dtf[new_col1] = list(cdata)
        if not type(self.__columns) == OrderedDict:
            self.__columns = OrderedDict()
        self.__columns[new_col1] = new_col2
        if not type(self.__indexes) == OrderedDict:
            self.__indexes = OrderedDict()
        labels1.append('_na')
        labels2.append(u'無回答')
        self.__indexes[new_col1] = OrderedDict(zip(labels1, labels2))
        return self.__dtf

    def table(self, gkey1, gkey2):
        dtfx = self.__cross_base(gkey1, gkey2)
        dtfx.columns = list(dtfx.columns)
        dtfs = pd.concat([dtfx, pd.DataFrame(dtfx.sum(axis=0),
            columns=[u'合計']).T]) # add the last row(sum: sum)
        dtfs = pd.concat([dtfs, pd.DataFrame(dtfs.sum(axis=1),
            columns=[u'合計'])], axis=1) # add the last column(sum: sum)
        dtfp1 = dtfs.T
        for col in dtfp1.columns:
            xsum = dtfp1[col][-1]
            if xsum == 0:
                xsum = 1
            dtfp1[col] = dtfp1[col] / xsum * 100.0
            if type(self.cround) == int:
                dtfp1[col] = dtfp1[col].round(self.cround)
        dtfp1 = dtfp1.T
        dtfp2 = dtfs.copy()
        for col in dtfp2.columns:
            xsum = dtfp2[col][-1]
            if xsum == 0:
                xsum = 1
            dtfp2[col] = dtfp2[col] / xsum * 100.0
            if type(self.cround) == int:
                dtfp2[col] = dtfp2[col].round(self.cround)
        nn = [gkey1, gkey2]
        if type(self.__columns) == OrderedDict:
            for i in [0, 1]:
                if nn[i] in self.__columns.keys():
                    nn[i] = self.__columns[nn[i]]
        dod = OrderedDict()
        dod['name'] = nn[0] + "/" + nn[1]
        dod['tbl1'] = dtfx
        dod['tbl2'] = dtfs
        dod['pct1'] = dtfp1
        dod['pct2'] = dtfp2
        return dod

    def chisq_test(self, gkey1, gkey2):
        dtfx = self.__cross_base(gkey1, gkey2)
        dd = chisq_test(dtfx)
        if not type(dd) == OrderedDict:
            return None
        nn = [gkey1, gkey2]
        if type(self.__columns) == OrderedDict:
            for i in [0, 1]:
                if nn[i] in self.__columns.keys():
                    nn[i] = self.__columns[nn[i]]
        dod = OrderedDict()
        dod['name'] = nn[0] + "/" + nn[1]
        dod.update(dd)
        return dod

    def groupby(self, cname, gkey):
        inames = self.__get_inames(gkey)
        if not inames is None:
            ilist = list(inames.keys())
        else:
            ilist = self.__unique(gkey)
        ser = self.__dtf[gkey] == self.__dtf[gkey]
        dtf = self.__dtf[ser]
        dgb = dtf.groupby(gkey)  # groupping by gkey
        dct = dict()
        for g_name, g_dtf in dgb:  # dgb to dict
            dct[g_name] = g_dtf
        dct = self.__key_str(dct)
        dod = OrderedDict()
        for idx in ilist:
            if not idx in dct.keys():
                continue
            if not cname in dct[idx].keys():
                continue
            ser = dct[idx][cname].dropna().reset_index(drop=True)
            key = idx
            if type(inames) == OrderedDict:
                key = inames[idx]
            dod[key] = ser
        return dod

    def crange(self, arg1, arg2=None, type=str):
        if self.__dtf is None:
            return None
        if arg2 is None:
            arg2 = arg1
        cl = list(self.__dtf.columns)
        dcl = dict()
        for i in range(len(cl)):
            dcl[cl[i]] = i
        flag = False
        if arg1 in cl and arg2 in cl:
            flag = True
        elif arg1 in dcl.values() and arg2 in dcl.values():
            arg1 = cl[arg1]
            arg2 = cl[arg2]
            flag = True
        if not flag:
            print2(u"Error: crange(), 引数 '%s' または '%s' が不適切!" \
                % (arg1, arg2))
            return None
        if dcl[arg1] > dcl[arg2]:
            w = arg1
            arg1 = arg2
            arg2 = w
        rng = []
        for i in range(len(cl)):
            val = cl[i]
            if type == int:
                val = i
            if cl[i] == arg2:
                rng.append(val)
                break
            if arg1 == cl[i]:
                rng.append(val)
                continue
            if len(rng) > 0:
                rng.append(val)
        return rng

    def irange(self, arg1, arg2=None, type=str):
        if self.__dtf is None:
            return None
        if arg2 is None:
            arg2 = arg1
        il = list(self.__dtf.index)
        dil = dict()
        for i in range(len(il)):
            dil[il[i]] = i
        flag = False
        if arg1 in il and arg2 in il:
            flag = True
        elif arg1 in dil.values() and arg2 in dil.values():
            arg1 = il[arg1]
            arg2 = il[arg2]
            flag = True
        if not flag:
            print2(u"Error: irange(), 引数 '%s' または '%s' が不適切!" \
                % (arg1, arg2))
            return None
        if dil[arg1] > dil[arg2]:
            w = arg1
            arg1 = arg2
            arg2 = w
        rng = []
        for i in range(len(il)):
            val = il[i]
            if type == int:
                val = i
            if il[i] == arg2:
                rng.append(val)
                break
            if arg1 == il[i]:
                rng.append(val)
                continue
            if len(rng) > 0:
                rng.append(val)
        return rng

    def var_test(self, cname, gkey, alternative="two-sided"):
        gg = self.groupby(cname, gkey)
        gglist = list(gg.keys())
        a = gg[gglist[0]]
        b = gg[gglist[1]]
        dd = var_test(a, b, alternative)
        dod = OrderedDict()
        dod['name'] = gglist[0] + "/" + gglist[1]
        dod.update(dd)
        return dod

    def t_test(self, cname, gkey, alternative="two-sided", \
            paired=False, equal_var=False):
        gg = self.groupby(cname, gkey)
        gglist = list(gg.keys())
        a = gg[gglist[0]]
        b = gg[gglist[1]]
        dd = t_test(a, b, alternative, paired, equal_var)
        dod = OrderedDict()
        dod['name'] = gglist[0] + "/" + gglist[1]
        dod.update(dd)
        return dod

    def ma_select(self, clist, value=1, name="ma_select"):
        ma_irng = self.__dtf.index.values
        dtfr = self.__dtf.loc[ma_irng, clist].T
        ary = list()
        for col in dtfr.columns:
            ser = dtfr[col].apply(lambda x: True if x == value \
                else np.NaN).dropna()
            ary.append(len(ser))
        ser = pd.Series(ary)
        ser.index = ma_irng
        if name in self.__dtf.columns:
            self.__dtf.drop(name, axis=1, inplace=True)
        self.__dtf[name] = ser
        return self.__dtf

    def ma_count(self, clist, asc=None, value=1):
        if self.__dtf is None:
            return None
        if not type(clist) == list:
            return None
        if not type(clist[0]) == list:
            clist = [clist]
        cnames = self.__get_cnames()
        dod = OrderedDict()
        dod_count = 0
        for rng in clist:
            ary = []
            for col in rng:
                ser = self.__dtf[col]
                ser = ser.apply(lambda x: True if x == value \
                    else np.NaN).dropna()
                ary.append(len(ser))
            dtfx = pd.DataFrame()
            dtfx['count'] = pd.Series(ary)
            dtfx.index = self.__to_str(rng)
            tsum = len(self.__dtf.index.values)
            ser = dtfx['count']
            dtfx['ratio'] = ser / tsum * 100.0
            if not cnames is None:
                dtfx.rename(index=cnames, inplace=True)
            if asc in [True, False]:
                dtfx = dtfx.sort_values('count', ascending=asc)
            wdtf = pd.DataFrame([[tsum, 100.0]], index=['_all'],
                columns=['count', 'ratio'])
            dtfx = dtfx.append(wdtf)
            if type(self.cround) == int:
                dtfx['ratio'] = dtfx['ratio'].round(self.cround)
            dtfx.rename(index={'_all': u'総数'}, inplace=True)
            dtfx.rename(columns=self.etoj3, inplace=True)
            dod[dod_count] = dtfx
            dod_count = dod_count + 1
        return dod

    def ma_table(self, rng, gkey, value=1):
        ser = self.__dtf[gkey] == self.__dtf[gkey]
        dtf = self.__dtf[ser]
        dgb = dtf.groupby(gkey)
        dct = OrderedDict()
        for g_name, g_dtf in dgb:
            dct[g_name] = g_dtf
        dct['_sum'] = dtf
        dct = self.__key_str(dct)
        tsum = []
        dtft2 = pd.DataFrame()
        for key in dct.keys():
            dtfw = dct[key]
            tsum.append(len(dtfw[gkey]))
            wcount = []
            for col in rng:
                ser = dtfw[col].apply(lambda x: \
                    True if x == value else np.NaN).dropna()
                wcount.append(len(ser))
            dtft2[key] = wcount
        dtft2.index = rng
        dtft2 = dtft2.T
        dtft2['_all'] = tsum
        dtft2.columns = self.__to_str(list(dtft2.columns))
        cnames = self.__get_cnames()
        if not cnames is None:
            dtft2.rename(columns=cnames, inplace=True)
        inames = self.__get_inames(gkey)
        if not inames is None:
            dtft2.rename(index=inames, inplace=True)
        dtft1 = dtft2.drop('_sum')
        dtft1 = dtft1.drop('_all', axis=1)
        dtft2.rename(columns={'_all': u'総数'}, inplace=True)
        dtft2.rename(index={'_sum': u'合計'}, inplace=True)
        dod = OrderedDict()
        dod['tbl1'] = dtft1
        dod['tbl2'] = dtft2
        dtfp1 = dtft2.T
        for col in dtfp1.columns:
            ser = dtfp1[col]
            ser = ser / ser[-1] * 100.0
            if type(self.cround) == int:
                ser = ser.round(self.cround)
            dtfp1[col] = ser
        dtfp1 = dtfp1.T
        dod['pct1'] = dtfp1
        dtfp2 = dtft2.copy()
        for col in dtfp2.columns:
            ser = dtfp2[col]
            ser = ser / ser[-1] * 100.0
            if type(self.cround) == int:
                ser = ser.round(self.cround)
            dtfp2[col] = ser
        dod['pct2'] = dtfp2
        return dod

    def ma_rel(self, clist, value=1):
        rng = clist
        ln = len(rng)
        dct = {}
        sers = []
        for col in rng:
            ser = self.__dtf[col]
            ser = ser.apply(lambda x: True if x == value else False)
            sers.append(ser)
        for i in range(ln):
            ser1 = sers[i]
            key = "%s" % [i,i]
            dct[key] = len(ser1[ser1])
            for j in range(i+1, ln):
                ser2 = sers[j]
                ser = ser1 & ser2
                ser = ser[ser]
                key = "%s" % [i,j]
                dct[key] = len(ser)
        tlist = list(range(ln))
        dtft = pd.DataFrame()
        for i in tlist:
           dtft[i] = tlist
        for key in dct.keys():
            i, j = eval(key)
            val = dct[key]
            if i == j:
                dtft.at[i, j] = val
            else:
                dtft.at[i, j] = val
                dtft.at[j, i] = val
        dtfp1 = pd.DataFrame()
        for i in tlist:
            dtfp1[i] = pd.Series(tlist).astype(float)
        for i in tlist:
            tsum = dtft.at[i, i] * 1.0
            for j in tlist:
                val = dtft.at[i, j] * 1.0
                val = val / tsum * 100.0
                dtfp1.at[i, j] = val
        dtfp2 = pd.DataFrame()
        for i in tlist:
            dtfp2[i] = pd.Series(tlist).astype(float)
        nsum = len(self.__dtf.index.values)
        for i in tlist:
            for j in tlist:
                val = dtft.at[i, j] * 1.0
                val = val / nsum * 100.0
                dtfp2.at[i, j] = val
        dod = OrderedDict()
        dod['tbl'] = dtft
        dod['pct1'] = dtfp1
        dod['pct2'] = dtfp2
        for key in dod.keys():
            dod[key].columns = rng
            dod[key].index = rng
            if key in ['pct1', 'pct2'] and type(self.cround) == int:
                dod[key] = dod[key].round(self.cround)
        dod['pct2'].columns.name = "N=%d" % nsum
        cnames = self.__get_cnames()
        if not cnames is None:
            for key in dod.keys():
                dod[key].rename(columns=cnames, inplace=True)
                dod[key].rename(index=cnames, inplace=True)
        return dod

